# Copyright 2018 Google LLC with subsequent modifications.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python2-compatible implementation for profilers."""

import atexit
import collections
import logging
import signal
import threading
import time
import timeit
from zprofile import builder

# Maximum stack frames to record.
_MAX_STACK_DEPTH = 128
_NANOS_PER_SEC = 1000 * 1000 * 1000

logger = logging.getLogger(__name__)


class WallProfiler(object):
  """Python2-compatible implementation for Wall time profiler.

  This Wall profiler avoids dependency on any native code by using
  Python signal module. It only profiles the main thread. It has the following
  limitations:

  1. A system call interrupted by a signal fails with EINTR. Python doesn't
     handle EINTR properly until Python 3.5. Python 3.4 handles it in some
     modules but not all, see https://www.python.org/dev/peps/pep-0475/. In
     earlier Python, when EINTR occurs, some system calls fail with Interrupted
     Error, such as socket.recv, and some fail silently, such as sleep (Python
     doc does mention that sleep can be terminated earlier than requested:
     https://docs.python.org/2/library/time.html#time.sleep). Python assumes
     that the application code takes the responsibility to handle EINTR properly
     (usually means retry in a loop). This profiler triggers signal so it can
     cause EINTR. The risk is reduced by requiring the operation system to
     restart the interrupted system call automatically. But not all system calls
     can be restarted by the OS. See "Interruption of system calls and library
     functions by signal handlers" in
     http://man7.org/linux/man-pages/man7/signal.7.html. Using this profiler
     with Python earlier than 3.5 requires that the application code handles
     EINTR properly.
  2. The Python-side signal handling, that this profiler uses, is capable of
     executing the signal handler on the main thread only. It achieves that by
     quickly acquiring and releasing GIL in a loop until the main thread gets to
     execute (see slide 23 of http://www.dabeaz.com/python/GIL.pdf). This comes
     at expense, especially on programs with large number of threads (which are
     not that common in Python which is good news). Also this means that the
     main thread must be in a state where it can get something executed - e.g.
     it won't if it waits on a thread join.
  3. The Python signal module in Python versions older than 3.6 doesn't handle
     signals properly. Briefly speaking, when a signal arrives, a global flag
     is_tripped is used to track whether PyErr_CheckSignals is already added to
     pending calls. PyErr_CheckSignals will clear this flag when it's called. In
     some race conditions, the flag is set to 1 but PyErr_CheckSignals is not
     added to the pending calls. This causes the program to no longer handle any
     signals. For example, Ctrl + C will not kill the process. The profiler
     triggers signal with high frequency, thus makes this problem more likely to
     happen. Based on experimentation, the problem is more likely to occur in
     Python 3, so the Wall profiler is not supported for Python versions
     (inclusive) 3 to 3.5. For Python 2, the problem theoretically can happen,
     but is much less likely to manifest. Users can use it at their own risk.
  """

  def __init__(self, period_ms=10):
    """Constructs the Wall time profiler.

    Args:
      period_ms: An integer specifying the sampling interval in milliseconds.
    """
    self._profile_type = 'wall'
    self._period_sec = float(period_ms) / 1000
    self._traces = collections.defaultdict(int)
    self._in_handler = False
    self._started = False
    self._last_sample_time = None
    self._trace_count = 0
    self._sample_time_lock = threading.RLock()

  def register_handler(self):
    """Registers the handler to the SIGALRM signal.

    This method must be called from the main thread. Attempting to call it from
    other threads will cause a ValueError exception.
    """
    signal.signal(signal.SIGALRM, self._handler)
    # Requires that the system restarts system calls interrupted by SIGALRM.
    # Not all calls can be restarted, see
    # http://man7.org/linux/man-pages/man7/signal.7.html.
    signal.siginterrupt(signal.SIGALRM, False)

    # Stop sending SIGALRM before the program exits. If SIGALRM is received
    # during the program exit, sometimes the program exits with non-zero code
    # 142. See b/133360821.
    atexit.register(signal.setitimer, signal.ITIMER_REAL, 0)

  def profile(self, profile_duration):
    """Profiles for the given duration.

    This function can be called from a non-main thread. It assumes
    register_handler has been called.

    Args:
      profile_duration: An float specifying the duration to profile in seconds.

    Returns:
      A bytes object containing gzip-compressed profile proto.
    """
    self._reset()

    duration_ns = int(_NANOS_PER_SEC * profile_duration)
    target_time = timeit.default_timer() + profile_duration

    self._start_profiling()
    # In Python 2, sleep can be interrupted by signal. Retries sleep until the
    # target time is reached.
    while profile_duration > 0:
      time.sleep(profile_duration)
      self._sample_time_lock.acquire()

      # Signal timer must be disabled before allocating memory. A fork call that
      # takes longer than the signal interval will enter an endless loop of
      # retrying interrupted clone system call:
      # http://lists.debian.org/debian-glibc/2010/03/msg00161.html. If that
      # happens, trying to allocation memory may hang waiting for memory lock.
      # Failing to disable the signal before allocating memory may cause a
      # deadlock.
      signal.setitimer(signal.ITIMER_REAL, 0)
      profile_duration = target_time - timeit.default_timer()
      signal.setitimer(signal.ITIMER_REAL, self._period_sec, self._period_sec)
      self._last_sample_time = None
      self._sample_time_lock.release()

    self._stop_profiling()

    return self._serialize_and_clear_traces(duration_ns)

  def _record_trace(self, frame):
    """Records the call stack trace of the given frame.

    Args:
      frame: A Frame object representing the leaf frame of the stack.

    Returns:
      A tuple of frames. The leaf frame is at position 0. A frame is a
      (function name, filename, line number) tuple.
    """
    depth = 0
    trace = []
    while frame is not None and depth < _MAX_STACK_DEPTH:
      frame_tuple = (frame.f_code.co_name, frame.f_code.co_filename,
                     frame.f_lineno)
      trace.append(frame_tuple)
      frame = frame.f_back
      depth += 1
    return tuple(trace)

  def _handler(self, unused_signum, frame):
    """Records the current call stack trace when signal received.

    In Python, signal can only occur between the atomic instructions of the
    Python interpreter. Since creating a string is "atomic" in Python sense,
    it's safe to copy strings in the signal handler. Also, signals are only
    handled by main thread. Simply using a bool flag to prevent reentry is
    fine.

    Args:
      frame: A Frame object representing the current stack frame.
    """

    # _started flags is used to ignore late signals that may be delivered after
    # the profiler has been stopped.
    if not self._started or self._in_handler:
      return

    self._in_handler = True
    trace = self._record_trace(frame)

    # Signal handler is only called when the execution returns to Python level
    # and when the main thread aquires the GIL. It's possible that multiple
    # signals occurred before the handler is called, for example, when Python
    # code calls into a long running C code such as gzip. The good news is that
    # we know that when the missed signal happened, the main thread Python level
    # stack is the same as the current one: if the main thread got a chance to
    # update the Python level stack, it already handled the signal. It's
    # appropriate to attribute the missed signals to the current stack.

    # Python signal handler is only called on main thread. This lock is a
    # reentrant lock which can be acquired again by the same thread without
    # blocking. We also prevented reentry of the handler function. So it's fine
    # to use this lock in the signal handler.
    self._sample_time_lock.acquire()
    now = timeit.default_timer()
    signal_tick_count = 1
    if self._last_sample_time is None:
      self._last_sample_time = now
    else:
      signal_tick_count = int((now - self._last_sample_time) / self._period_sec)
      signal_tick_count = max(1, signal_tick_count)
      self._last_sample_time += self._period_sec * signal_tick_count
    self._sample_time_lock.release()

    self._traces[trace] += signal_tick_count
    self._trace_count += signal_tick_count
    self._in_handler = False

  def _start_profiling(self):
    self._started = True
    signal.setitimer(signal.ITIMER_REAL, self._period_sec, self._period_sec)

  def _stop_profiling(self):
    """Stops timer and waits for the last handler to finish."""
    signal.setitimer(signal.ITIMER_REAL, 0)
    self._started = False

    # Waits for the last signal handler to finish.
    count = 0
    while self._in_handler:
      if count % 1000 == 0:
        logger.info('Wait for the last signal handler to finish')
      count += 1
      # This also releases the GIL to allow the main thread to finish handling.
      time.sleep(0.01)

  def _serialize_and_clear_traces(self, duration_ns):
    period_ns = int(self._period_sec * _NANOS_PER_SEC)
    unknown_trace_count = int(duration_ns / period_ns) - self._trace_count
    if unknown_trace_count > 0:
      self._traces[(('unknown', 'unknown', 0),)] = unknown_trace_count
    profile_builder = builder.Builder()
    profile_builder.populate_profile(self._traces, self._profile_type,
                                     'nanoseconds', period_ns, duration_ns)
    self._reset()
    return profile_builder.emit()

  def _reset(self):
    self._traces = collections.defaultdict(int)
    self._last_sample_time = None
    self._trace_count = 0



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: profile.proto
# pylint:skip-file

import sys
_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='profile.proto',
    package='perftools.profiles',
    syntax='proto3',
    serialized_pb=_b(
        '\n\rprofile.proto\x12\x12perftools.profiles\"\xd5\x03\n\x07Profile\x12\x32\n\x0bsample_type\x18\x01 \x03(\x0b\x32\x1d.perftools.profiles.ValueType\x12*\n\x06sample\x18\x02 \x03(\x0b\x32\x1a.perftools.profiles.Sample\x12,\n\x07mapping\x18\x03 \x03(\x0b\x32\x1b.perftools.profiles.Mapping\x12.\n\x08location\x18\x04 \x03(\x0b\x32\x1c.perftools.profiles.Location\x12.\n\x08\x66unction\x18\x05 \x03(\x0b\x32\x1c.perftools.profiles.Function\x12\x14\n\x0cstring_table\x18\x06 \x03(\t\x12\x13\n\x0b\x64rop_frames\x18\x07 \x01(\x03\x12\x13\n\x0bkeep_frames\x18\x08 \x01(\x03\x12\x12\n\ntime_nanos\x18\t \x01(\x03\x12\x16\n\x0e\x64uration_nanos\x18\n \x01(\x03\x12\x32\n\x0bperiod_type\x18\x0b \x01(\x0b\x32\x1d.perftools.profiles.ValueType\x12\x0e\n\x06period\x18\x0c \x01(\x03\x12\x0f\n\x07\x63omment\x18\r \x03(\x03\x12\x1b\n\x13\x64\x65\x66\x61ult_sample_type\x18\x0e \x01(\x03\"\'\n\tValueType\x12\x0c\n\x04type\x18\x01 \x01(\x03\x12\x0c\n\x04unit\x18\x02 \x01(\x03\"V\n\x06Sample\x12\x13\n\x0blocation_id\x18\x01 \x03(\x04\x12\r\n\x05value\x18\x02 \x03(\x03\x12(\n\x05label\x18\x03 \x03(\x0b\x32\x19.perftools.profiles.Label\"@\n\x05Label\x12\x0b\n\x03key\x18\x01 \x01(\x03\x12\x0b\n\x03str\x18\x02 \x01(\x03\x12\x0b\n\x03num\x18\x03 \x01(\x03\x12\x10\n\x08num_unit\x18\x04 \x01(\x03\"\xdd\x01\n\x07Mapping\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x14\n\x0cmemory_start\x18\x02 \x01(\x04\x12\x14\n\x0cmemory_limit\x18\x03 \x01(\x04\x12\x13\n\x0b\x66ile_offset\x18\x04 \x01(\x04\x12\x10\n\x08\x66ilename\x18\x05 \x01(\x03\x12\x10\n\x08\x62uild_id\x18\x06 \x01(\x03\x12\x15\n\rhas_functions\x18\x07 \x01(\x08\x12\x15\n\rhas_filenames\x18\x08 \x01(\x08\x12\x18\n\x10has_line_numbers\x18\t \x01(\x08\x12\x19\n\x11has_inline_frames\x18\n \x01(\x08\"v\n\x08Location\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x12\n\nmapping_id\x18\x02 \x01(\x04\x12\x0f\n\x07\x61\x64\x64ress\x18\x03 \x01(\x04\x12&\n\x04line\x18\x04 \x03(\x0b\x32\x18.perftools.profiles.Line\x12\x11\n\tis_folded\x18\x05 \x01(\x08\")\n\x04Line\x12\x13\n\x0b\x66unction_id\x18\x01 \x01(\x04\x12\x0c\n\x04line\x18\x02 \x01(\x03\"_\n\x08\x46unction\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x01(\x03\x12\x13\n\x0bsystem_name\x18\x03 \x01(\x03\x12\x10\n\x08\x66ilename\x18\x04 \x01(\x03\x12\x12\n\nstart_line\x18\x05 \x01(\x03\x42-\n\x1d\x63om.google.perftools.profilesB\x0cProfileProtob\x06proto3'
    ))
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_PROFILE = _descriptor.Descriptor(
    name='Profile',
    full_name='perftools.profiles.Profile',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='sample_type',
            full_name='perftools.profiles.Profile.sample_type',
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='sample',
            full_name='perftools.profiles.Profile.sample',
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='mapping',
            full_name='perftools.profiles.Profile.mapping',
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='location',
            full_name='perftools.profiles.Profile.location',
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='function',
            full_name='perftools.profiles.Profile.function',
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='string_table',
            full_name='perftools.profiles.Profile.string_table',
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='drop_frames',
            full_name='perftools.profiles.Profile.drop_frames',
            index=6,
            number=7,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='keep_frames',
            full_name='perftools.profiles.Profile.keep_frames',
            index=7,
            number=8,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='time_nanos',
            full_name='perftools.profiles.Profile.time_nanos',
            index=8,
            number=9,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='duration_nanos',
            full_name='perftools.profiles.Profile.duration_nanos',
            index=9,
            number=10,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='period_type',
            full_name='perftools.profiles.Profile.period_type',
            index=10,
            number=11,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='period',
            full_name='perftools.profiles.Profile.period',
            index=11,
            number=12,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='comment',
            full_name='perftools.profiles.Profile.comment',
            index=12,
            number=13,
            type=3,
            cpp_type=2,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='default_sample_type',
            full_name='perftools.profiles.Profile.default_sample_type',
            index=13,
            number=14,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=38,
    serialized_end=507,
)

_VALUETYPE = _descriptor.Descriptor(
    name='ValueType',
    full_name='perftools.profiles.ValueType',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='type',
            full_name='perftools.profiles.ValueType.type',
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='unit',
            full_name='perftools.profiles.ValueType.unit',
            index=1,
            number=2,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=509,
    serialized_end=548,
)

_SAMPLE = _descriptor.Descriptor(
    name='Sample',
    full_name='perftools.profiles.Sample',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='location_id',
            full_name='perftools.profiles.Sample.location_id',
            index=0,
            number=1,
            type=4,
            cpp_type=4,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='value',
            full_name='perftools.profiles.Sample.value',
            index=1,
            number=2,
            type=3,
            cpp_type=2,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='label',
            full_name='perftools.profiles.Sample.label',
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=550,
    serialized_end=636,
)

_LABEL = _descriptor.Descriptor(
    name='Label',
    full_name='perftools.profiles.Label',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='key',
            full_name='perftools.profiles.Label.key',
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='str',
            full_name='perftools.profiles.Label.str',
            index=1,
            number=2,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='num',
            full_name='perftools.profiles.Label.num',
            index=2,
            number=3,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='num_unit',
            full_name='perftools.profiles.Label.num_unit',
            index=3,
            number=4,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=638,
    serialized_end=702,
)

_MAPPING = _descriptor.Descriptor(
    name='Mapping',
    full_name='perftools.profiles.Mapping',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='id',
            full_name='perftools.profiles.Mapping.id',
            index=0,
            number=1,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='memory_start',
            full_name='perftools.profiles.Mapping.memory_start',
            index=1,
            number=2,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='memory_limit',
            full_name='perftools.profiles.Mapping.memory_limit',
            index=2,
            number=3,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='file_offset',
            full_name='perftools.profiles.Mapping.file_offset',
            index=3,
            number=4,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='filename',
            full_name='perftools.profiles.Mapping.filename',
            index=4,
            number=5,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='build_id',
            full_name='perftools.profiles.Mapping.build_id',
            index=5,
            number=6,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='has_functions',
            full_name='perftools.profiles.Mapping.has_functions',
            index=6,
            number=7,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='has_filenames',
            full_name='perftools.profiles.Mapping.has_filenames',
            index=7,
            number=8,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='has_line_numbers',
            full_name='perftools.profiles.Mapping.has_line_numbers',
            index=8,
            number=9,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='has_inline_frames',
            full_name='perftools.profiles.Mapping.has_inline_frames',
            index=9,
            number=10,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=705,
    serialized_end=926,
)

_LOCATION = _descriptor.Descriptor(
    name='Location',
    full_name='perftools.profiles.Location',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='id',
            full_name='perftools.profiles.Location.id',
            index=0,
            number=1,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='mapping_id',
            full_name='perftools.profiles.Location.mapping_id',
            index=1,
            number=2,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='address',
            full_name='perftools.profiles.Location.address',
            index=2,
            number=3,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='line',
            full_name='perftools.profiles.Location.line',
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='is_folded',
            full_name='perftools.profiles.Location.is_folded',
            index=4,
            number=5,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=928,
    serialized_end=1046,
)

_LINE = _descriptor.Descriptor(
    name='Line',
    full_name='perftools.profiles.Line',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='function_id',
            full_name='perftools.profiles.Line.function_id',
            index=0,
            number=1,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='line',
            full_name='perftools.profiles.Line.line',
            index=1,
            number=2,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=1048,
    serialized_end=1089,
)

_FUNCTION = _descriptor.Descriptor(
    name='Function',
    full_name='perftools.profiles.Function',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='id',
            full_name='perftools.profiles.Function.id',
            index=0,
            number=1,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='name',
            full_name='perftools.profiles.Function.name',
            index=1,
            number=2,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='system_name',
            full_name='perftools.profiles.Function.system_name',
            index=2,
            number=3,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='filename',
            full_name='perftools.profiles.Function.filename',
            index=3,
            number=4,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='start_line',
            full_name='perftools.profiles.Function.start_line',
            index=4,
            number=5,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=1091,
    serialized_end=1186,
)

_PROFILE.fields_by_name['sample_type'].message_type = _VALUETYPE
_PROFILE.fields_by_name['sample'].message_type = _SAMPLE
_PROFILE.fields_by_name['mapping'].message_type = _MAPPING
_PROFILE.fields_by_name['location'].message_type = _LOCATION
_PROFILE.fields_by_name['function'].message_type = _FUNCTION
_PROFILE.fields_by_name['period_type'].message_type = _VALUETYPE
_SAMPLE.fields_by_name['label'].message_type = _LABEL
_LOCATION.fields_by_name['line'].message_type = _LINE
DESCRIPTOR.message_types_by_name['Profile'] = _PROFILE
DESCRIPTOR.message_types_by_name['ValueType'] = _VALUETYPE
DESCRIPTOR.message_types_by_name['Sample'] = _SAMPLE
DESCRIPTOR.message_types_by_name['Label'] = _LABEL
DESCRIPTOR.message_types_by_name['Mapping'] = _MAPPING
DESCRIPTOR.message_types_by_name['Location'] = _LOCATION
DESCRIPTOR.message_types_by_name['Line'] = _LINE
DESCRIPTOR.message_types_by_name['Function'] = _FUNCTION

Profile = _reflection.GeneratedProtocolMessageType(
    'Profile',
    (_message.Message,),
    dict(
        DESCRIPTOR=_PROFILE,
        __module__='profile_pb2'
        # @@protoc_insertion_point(class_scope:perftools.profiles.Profile)
    ))
_sym_db.RegisterMessage(Profile)

ValueType = _reflection.GeneratedProtocolMessageType(
    'ValueType',
    (_message.Message,),
    dict(
        DESCRIPTOR=_VALUETYPE,
        __module__='profile_pb2'
        # @@protoc_insertion_point(class_scope:perftools.profiles.ValueType)
    ))
_sym_db.RegisterMessage(ValueType)

Sample = _reflection.GeneratedProtocolMessageType(
    'Sample',
    (_message.Message,),
    dict(
        DESCRIPTOR=_SAMPLE,
        __module__='profile_pb2'
        # @@protoc_insertion_point(class_scope:perftools.profiles.Sample)
    ))
_sym_db.RegisterMessage(Sample)

Label = _reflection.GeneratedProtocolMessageType(
    'Label',
    (_message.Message,),
    dict(
        DESCRIPTOR=_LABEL,
        __module__='profile_pb2'
        # @@protoc_insertion_point(class_scope:perftools.profiles.Label)
    ))
_sym_db.RegisterMessage(Label)

Mapping = _reflection.GeneratedProtocolMessageType(
    'Mapping',
    (_message.Message,),
    dict(
        DESCRIPTOR=_MAPPING,
        __module__='profile_pb2'
        # @@protoc_insertion_point(class_scope:perftools.profiles.Mapping)
    ))
_sym_db.RegisterMessage(Mapping)

Location = _reflection.GeneratedProtocolMessageType(
    'Location',
    (_message.Message,),
    dict(
        DESCRIPTOR=_LOCATION,
        __module__='profile_pb2'
        # @@protoc_insertion_point(class_scope:perftools.profiles.Location)
    ))
_sym_db.RegisterMessage(Location)

Line = _reflection.GeneratedProtocolMessageType(
    'Line',
    (_message.Message,),
    dict(
        DESCRIPTOR=_LINE,
        __module__='profile_pb2'
        # @@protoc_insertion_point(class_scope:perftools.profiles.Line)
    ))
_sym_db.RegisterMessage(Line)

Function = _reflection.GeneratedProtocolMessageType(
    'Function',
    (_message.Message,),
    dict(
        DESCRIPTOR=_FUNCTION,
        __module__='profile_pb2'
        # @@protoc_insertion_point(class_scope:perftools.profiles.Function)
    ))
_sym_db.RegisterMessage(Function)

DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(
    descriptor_pb2.FileOptions(),
    _b('\n\035com.google.perftools.profilesB\014ProfileProto'))
# @@protoc_insertion_point(module_scope)



# Copyright 2018 Google LLC with subsequent modifications.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CPU time profiler."""

import logging
from zprofile import _profiler
from zprofile import builder

_NANOS_PER_SEC = 1000 * 1000 * 1000

logger = logging.getLogger(__name__)


class CPUProfiler(object):
  """CPU time profiler.

  The profiler collects CPU time usage data and builds the data as
  a gzip-compressed profile proto.
  """

  def __init__(self, period_ms=10):
    """Constructs the CPU time profiler.

    Args:
      period_ms: An optional integer specifying the sampling interval in
        milliseconds. Defaults to 10.
    """
    self._profile_type = 'CPU'
    self._period_ms = period_ms

  def profile(self, duration_secs):
    """Profiles the CPU time usage for the given duration.

    Args:
      duration_secs: An float specifying the duration to profile in seconds.

    Returns:
      A bytes object containing gzip-compressed profile proto.
    """
    duration_ns = int(_NANOS_PER_SEC * duration_secs)
    traces = self._profile(duration_ns)
    return self._build_profile(duration_ns, traces)

  def _profile(self, duration_ns):
    return _profiler.profile_cpu(duration_ns, self._period_ms)

  def _build_profile(self, duration_ns, traces):
    profile_builder = builder.Builder()
    profile_builder.populate_profile(traces, self._profile_type, 'nanoseconds',
                                     self._period_ms * 1000 * 1000, duration_ns)
    return profile_builder.emit()



# Copyright 2018 Google LLC with subsequent modifications.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Builds the profile proto from call stack traces."""

import collections
import gzip
import io
from zprofile import profile_pb2

Func = collections.namedtuple('Func', ['name', 'filename'])
Loc = collections.namedtuple('Loc', ['func_id', 'line_number'])


class Builder(object):
  """Builds the profile proto from call stack traces."""

  def __init__(self):
    self._profile = profile_pb2.Profile()
    self._function_map = {}
    self._location_map = {}
    self._string_map = {}
    # string_table[0] in the profile proto must be an empty string.
    self._string_id('')

  def populate_profile(self, traces, profile_type, period_unit, period,
                       duration_ns):
    """Populates call stack traces into a profile proto.

    Args:
      traces: A map mapping a trace to its count. A trace is a sequence of
        frames. The leaf frame is at trace[0]. A frame is represented as a tuple
        of (function name, filename, line number).
      profile_type: A string specifying the profile type, e.g 'CPU' or 'WALL'.
        See https://github.com/google/pprof/blob/master/proto/profile.proto for
        possible profile types.
      period_unit: A string specifying the measurement unit of the sampling
        period, e.g 'nanoseconds'.
      period: An integer specifying the interval between sampled occurrences.
        The measurement unit is specified by the period_unit argument.
      duration_ns: An integer specifying the profiling duration in nanoseconds.
    """
    self._profile.period_type.type = self._string_id(profile_type)
    self._profile.period_type.unit = self._string_id(period_unit)
    self._profile.period = period
    self._profile.duration_nanos = duration_ns
    type1 = self._profile.sample_type.add()
    type1.type = self._string_id('sample')
    type1.unit = self._string_id('count')
    type2 = self._profile.sample_type.add()
    type2.type = self._string_id(profile_type)
    type2.unit = self._string_id(period_unit)

    for trace, count in traces.items():
      sample = self._profile.sample.add()
      sample.value.append(count)
      sample.value.append(period * count)
      for frame in trace:
        # TODO: try to use named tuple for frame if it doesn't over
        # complicate the native profiler.
        func_id = self._function_id(frame[0], frame[1])
        location_id = self._location_id(func_id, frame[2])
        sample.location_id.append(location_id)

  def emit(self):
    """Returns the profile in gzip-compressed profile proto format."""
    profile = self._profile.SerializeToString()
    out = io.BytesIO()
    with gzip.GzipFile(fileobj=out, mode='wb') as f:
      f.write(profile)
    return out.getvalue()

  def _function_id(self, name, filename):
    """Finds the function ID in the proto, adds the function if not yet exists.

    Args:
      name: A string representing the function name.
      filename: A string representing the file name.

    Returns:
      An integer representing the unique ID of the function in the profile
      proto.
    """
    name_id = self._string_id(name)
    filename_id = self._string_id(filename)
    func = Func(name_id, filename_id)

    func_id = self._function_map.get(func)
    if func_id is None:
      # Function ID in profile proto must not be zero.
      func_id = len(self._function_map) + 1
      self._function_map[func] = func_id
      function = self._profile.function.add()
      function.name = name_id
      function.filename = filename_id
      function.id = func_id
    return func_id

  def _location_id(self, func_id, line_number):
    """Finds the location ID in the proto, adds the location if not yet exists.

    Args:
      func_id: An integer representing the ID of the corresponding function in
        the profile proto.
      line_number: An integer representing the line number in the source code.

    Returns:
      An integer representing the unique ID of the location in the profile
      proto.
    """
    loc = Loc(func_id=func_id, line_number=line_number)

    location_id = self._location_map.get(loc)
    if location_id is None:
      # Location ID in profile proto must not be zero.
      location_id = len(self._location_map) + 1
      self._location_map[loc] = location_id
      location = self._profile.location.add()
      location.id = location_id
      line = location.line.add()
      line.line = line_number
      line.function_id = func_id
    return location_id

  def _string_id(self, value):
    """Finds the string ID in the proto, adds the string if not yet exists."""
    string_id = self._string_map.get(value)
    if string_id is None:
      string_id = len(self._string_map)
      self._string_map[value] = string_id
      self._profile.string_table.append(value)
    return string_id



# Copyright 2018 Google LLC with subsequent modifications.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cloud Profiler Python agent packaging script."""

from __future__ import print_function

import glob
import re
import sys
from setuptools import Extension
from setuptools import setup

install_requires = [
    'protobuf',
]

ext_module = [
    Extension(
        'zprofile._profiler',
        sources=glob.glob('zprofile/src/*.cc'),
        include_dirs=['zprofile/src'],
        language='c++',
        extra_compile_args=['-std=c++11'],
        extra_link_args=[
            '-std=c++11',
            '-static-libstdc++',
            # While libgcc_s.so.1 is pretty much always installed by default
            # for non-Alpine linux, it is not installed by default in Alpine.
            # So, to support Alpine, we will always statically link "libgcc"
            # package. We could alternatively require users to install the
            # "libgcc" package, but the static linkage seems less
            # invasive.
            '-static-libgcc'
        ])
]

if not (sys.platform.startswith('linux') or sys.platform.startswith('darwin')):
  print(
      sys.platform, 'is not a supported operating system.\n'
      'Profiler Python agent modules will be installed but will not '
      'be functional. Refer to the documentation for a list of '
      'supported operating systems.\n')
  ext_module = []


def get_version():
  """Read the version from __version__.py."""

  with open('zprofile/__version__.py') as fp:
    # Do not handle exceptions from open() so setup will fail when it cannot
    # open the file
    line = fp.read()
    version = re.search(r"^__version__ = '([0-9]+\.[0-9]+(\.[0-9]+)?-?.*)'",
                        line, re.M)
    if version:
      return version.group(1)

  raise RuntimeError(
      'Cannot determine version from zprofile/__init__.py.')


setup(
    name='zprofile',
    description='Statistical CPU and wall-clock profilers for Python',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/timpalpant/zprofile',
    author='Timothy Palpant',
    version=get_version(),
    install_requires=install_requires,
    setup_requires=['wheel'],
    packages=['zprofile'],
    ext_modules=ext_module,
    license='Apache License, Version 2.0',
    keywords='cloud profiler',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ])
